package Source::Repository::Mass;

use strict;
use warnings;
use Cwd;
use Carp;
use Getopt::Long;
use File::Basename;
use File::Path qw(make_path remove_tree);

use Source::Shared::Utils::GlobList qw/_line2array _expand_shell_glob/;
use Source::Shared::CLI 0.009;
use Source::Shared::ResourceContainer;
use Source::Shared::Resource::Debug;
use Source::Shared::Resource::Verbose;
use Source::Package::Pair;
use Source::Package::Comparator;
use Source::Package::Comparators;
use Source::Repository::Mass::NameFilter;
use Source::Repository::Mass::PkgFilter;
use Source::Repository 0.406;
use Source::Repository::RPM::NoEpoch;
use Source::Repository::Matcher;

require Exporter;

our $VERSION='0.448';

our @ISA = qw(Exporter Source::Shared::CLI);
our @EXPORT = qw/$DESTPATH $ORIGINPATH/;
our ($DESTPATH, $ORIGINPATH);

our $comparator_class='Source::Package::Comparator';
our $matcher_class='Source::Repository::Matcher';
our $origin_class='Source::Repository::RPM::NoEpoch';
our $dest_class='Source::Repository::RPM::NoEpoch';
our $origin_reponame='Origin repository';
our $dest_reponame='Destination repository';

our $epoch_comparator=\&Source::Package::Comparators::null_epoch_comparator;
our $version_comparator=\&Source::Package::Comparators::pure_rpm_version_comparator;
our $release_comparator=\&Source::Package::Comparators::pure_rpm_release_comparator;

our @ORIGIN_DEFAULT_OPTS;
our @DEST_DEFAULT_OPTS;

our $verbose=0;
our $strategy = 'all';

my ($opt_transaction, $opt_taskfile, $opt_all, $opt_exclude_namelist, $opt_destmode);
my ($opt_origin_namelist, $opt_dest_namelist, $opt_namelist);
my ($opt_origin_repolist, $opt_dest_repolist);
my ($opt_origin_default, $opt_dest_default);
my (@opt_origin_path, @opt_dest_path);
my ($opt_origin_cachedir, $opt_dest_cachedir);

our ($opt_skip_on_error);

our %STRATEGYMAP=(
    # quick synonyms
    'force' => \&origin2dest,
    'f' => \&origin2dest,
    'a' => \&origin2dest,
    'm' => \&originskipdest,
    'o' => \&destskiporigin,
    'u' => \&destup,
    'n' => \&newer,

    'all' => \&origin2dest,
    'missing' => \&originskipdest,
    'nodest' => \&originskipdest, # alias for missing
    'noorigin' => \&destskiporigin, # alias for orphaned
    'orphaned' => \&destskiporigin,
    'destup' => \&destup,
    'destnotnewer' => \&destnotnewer,
    'newer' => \&newer,
    'newolder' => \&newolder,
    'notolder' => \&notolder,
    'older' => \&older,
    'both' => \&both,
    # S(ilent) names
    'allS' => \&origin2destS,
    'missingS' => \&originskipdestS,
    'nodestS' => \&originskipdestS, # alias for missing
    'nooriginS' => \&destskiporiginS, # alias for orphaned
    'orphanedS' => \&destskiporiginS,
    'destupS' => \&destupS,
    'destnotnewerS' => \&destnotnewerS,
    'newerS' => \&newerS,
    'olderS' => \&olderS,
    'bothS' => \&bothS,

    # old names
    'origin2dest' => \&origin2dest,
    'originskipdest' => \&originskipdest,
    'destskiporigin' => \&destskiporigin,

    # verup, relup
    'verup' => \&verup,
    'relup' => \&relup,
    'same' => \&same,

    # print help
    'help' => \&help_strategy,
    );

our %STRATEGYDESCR=(
    'all' => 'any package',
    'missing' => 'packages in origin, missing in dest', 
    'orphaned' => 'packages in dest, missing in origin',
    'destup' => 'packages in dest, older than origin (i.e. need update)',
    'destnotnewer' => 'packages in dest, not newer than origin',
    'newer' => 'packages in origin, new and newer than dest',
    'newolder' => 'packages in origin, new and older than dest',
    'notolder' => 'packages in origin, new, newer or the same but not older than dest',
    'older' => 'packages in origin, older than dest',
    'both' => 'packages in origin, present in dest',
    'relup' => 'destup, limited release update',
    'verup' => 'destup, limited to version update',
    'same' => 'packages with same version and release up to a comparator',

    );

our @LONGOPT = (
	"all"	=> \$opt_all,
	"dest=s"	=> \$opt_dest_default,
	"dest-cachedir=s"	=> \$opt_dest_cachedir,
	"dest-namelist=s"	=> \$opt_dest_namelist,
	"dest-repolist=s"	=> \$opt_dest_repolist,
	"dest-path=s"		=> \@opt_dest_path,
	"exclude-namelist=s"	=> \$opt_exclude_namelist,
	"input-dest"	=> sub {$opt_destmode=1},
	"input-origin"	=> sub {$opt_destmode=0},
	"origin=s"	=> \$opt_origin_default,
	"origin-cachedir=s"	=> \$opt_origin_cachedir,
	"origin-namelist=s"	=> \$opt_origin_namelist,
	"origin-repolist=s"	=> \$opt_origin_repolist,
	"origin-path=s"		=> \@opt_origin_path,
	"skip-on-error"	=> \$opt_skip_on_error,
	"strategy=s"	=> \$strategy,
	"transaction"	=> \$opt_transaction,
	"tasklist=s"	=> \$opt_taskfile,
	"namelist=s"	=> \$opt_namelist,
 );

our @EXT_LONGOPT_CLASSES;

my @built_in_containers=(
    'Source::Shared::ResourceContainer', 
    'Source::Repository::Matcher', 
    'Source::Repository::Mass::NameFilter',
    'Source::Repository::Mass::PkgFilter',
);

sub cli_container_ref {
    return [@built_in_containers,@EXT_LONGOPT_CLASSES];
}

# TODO!!! too lacking!!!
sub print_longopt {
    print "
Mass processing options:
	--origin	/path1/to/foreign|origin/src/RPMS:/other/path2
	--dest	/path1/to/altlinux|dest/src/RPMS:/other/path1:/other/path2
	--skip-on-error	ignore individual package failures
	--input-origin	input names belong to origin repo (default)
	--input-dest	input names belong to dest repo
	--strategy 'strategy'	of mass operations (default is $strategy).
";&_print_strategies();
}

sub _print_strategies {
    print "	Available strategies: 
";
    my %main_name=map {$STRATEGYMAP{$_}=>$_} keys (%STRATEGYDESCR);
    my %synonym=map {$_=>[]} keys (%STRATEGYDESCR);
    foreach my $sname (sort keys(%STRATEGYMAP)) {
	my $description=$STRATEGYDESCR{$sname};
	my $mname=$main_name{$STRATEGYMAP{$sname}};
	push@{$synonym{$mname}}, $sname if $mname and not defined $description;
    }
    foreach my $sname (sort keys(%STRATEGYMAP)) {
	my $description=$STRATEGYDESCR{$sname};
	if (defined $description) {
	    print "\t\t$sname\t- $description\n";
	    my @synonyms = @{$synonym{$sname}};
	    print "\t\t\t  Synonyms: ".join(',',@synonyms)."\n" if @synonyms;
	}
    }
}


sub new {
    my $class = shift;
    my $self = {
	# replace given name with $pkg->name query (can be time consuming)
	-use_pkg_name => 0,
	'-originopts' => [],
	'-destopts' => [],
	# used in autorepo-mail-mainrepo-older
	    '-packages' => [],
	# used in autorepo-purge
	# -opt_all_names_def_repo => 0
	@_
    };
    bless $self, $class;
    $self->get_and_process_cli_options();
    print STDERR "Mass:usage: EXT_LONGOPT_CLASSES: ",join(', ',@EXT_LONGOPT_CLASSES),"\n" if $verbose and $verbose>6;

    my (@defrepo_opts, @origin_opts, @dest_opts);
    push @defrepo_opts, -namelist => $opt_namelist if $opt_namelist;
    push @origin_opts, -namelist => $opt_origin_namelist if $opt_origin_namelist;
    push @dest_opts, -namelist => $opt_dest_namelist if $opt_dest_namelist;
    push @origin_opts, -repolist => $opt_origin_repolist if $opt_origin_repolist;
    push @dest_opts, -repolist => $opt_dest_repolist if $opt_dest_repolist;
    push @origin_opts, -cachedir => $opt_origin_cachedir if $opt_origin_cachedir;
    push @dest_opts, -cachedir => $opt_dest_cachedir if $opt_dest_cachedir;
    push @origin_opts, -path => \@opt_origin_path if @opt_origin_path;
    push @dest_opts, -path => \@opt_dest_path if @opt_dest_path;
    push @origin_opts, -default_opt => $opt_origin_default if $opt_origin_default;
    push @dest_opts, -default_opt => $opt_dest_default if $opt_dest_default;
    push @origin_opts, -default_opt => $ORIGINPATH if $ORIGINPATH and not @opt_origin_path and not $opt_origin_repolist and not $opt_origin_default;
    push @dest_opts, -default_opt => $DESTPATH if $DESTPATH and not @opt_dest_path and not $opt_dest_repolist and not $opt_dest_default;

    eval {
	$self->{-originrepo}=$origin_class->create_repo_chain_from_cli(
	    -reponame => $origin_reponame,
	    @ORIGIN_DEFAULT_OPTS,
	    $opt_destmode ? () : @defrepo_opts,
	    @origin_opts,
	    @{$self->{-originopts}},
	    );
    };
    die "$@Mass.pm: Can't find the origin repository. Please, set the origin repository explicitly.\n" if $@;
    eval {
	$self->{-destrepo}=$dest_class->create_repo_chain_from_cli(
	    -reponame => $dest_reponame,
	    @DEST_DEFAULT_OPTS,
	    $opt_destmode ? @defrepo_opts : (),
	    @dest_opts,
	    @{$self->{-destopts}},
	    );
    };
    die "$@Mass.pm: Can't find the destination repository. Please, set the destination repository explicitly.\n" if $@;

    $self->{-destmode}=$opt_destmode;
    my $def_repo = $opt_destmode ? $self->{-destrepo} : $self->{-originrepo};
    if ($opt_taskfile) {
	open (my $fn, '<', $opt_taskfile) or die "can't open file $opt_taskfile\n";
	my $def_repo_all_names_ref=$def_repo->all_names_ref;
	while (my $line=<$fn>) {
	    my @list = map{&_expand_shell_glob($_,$def_repo_all_names_ref)} &_line2array($line);
	    next if scalar @list == 0;
	    if (scalar @list == 1) {
		push @{$self->{'-packages'}}, @list;
	    } elsif ($list[0] eq '@FOREACH:') {
		shift @list;
		push @{$self->{'-packages'}}, @list;
	    } else {
		# transaction
		push @{$self->{'-packages'}}, [@list];
	    }
	}
	close($fn);
    }

    my @in_args = @ARGV;
    push @in_args, @{$def_repo->all_names_ref()} if $opt_all or $self->{-opt_all_names_def_repo};
    if ($opt_exclude_namelist) {
	my %exclude;
	open (my $fn, '<', $opt_exclude_namelist) or die "can't open file $opt_exclude_namelist\n";
	while (my $line=<$fn>) {
	    next if $line=~/^\s*#/;
	    my @list = split(/\s+/,$line);
	    $exclude{$list[0]}=1;
	}
	close($fn);
	@in_args=grep {!$exclude{$_}} @in_args;
    }
    push @{$self->{'-packages'}}, $opt_transaction ? [@in_args] : @in_args;
    $self->{-comparator}=$comparator_class->new(
	-epoch_comparator => $epoch_comparator,
	-version_comparator=>$version_comparator,
	-release_comparator=>$release_comparator);
    if ($verbose>8) {
	print STDERR "DEBUG:Mass: ORIGINPATH=$ORIGINPATH\n";
	print STDERR "DEBUG:Mass: DESTPATH=$DESTPATH\n";
	print STDERR "DEBUG:Mass: -packages=",join(',',@{$self->{'-packages'}}),"\n";
	# as for now, used only nere
	&Source::Repository::Mass::NameFilter::debug();
	&Source::Repository::Mass::PkgFilter::debug();
    }
    #============= new =========
    return $self;
}

sub originrepo {
    my ($self)=@_;
    return $self->{-originrepo};
}

sub destrepo {
    my ($self)=@_;
    return $self->{-destrepo};
}

sub comparator {
    my ($self)=@_;
    return $self->{-comparator};
}

sub process {
    my ($self)=@_;
    foreach my $nameobj (@{$self->{'-packages'}}) {
	if (ref $nameobj) {
	    $self->_process_transaction($nameobj);
	} else {
	    $self->_process_name($nameobj);
	}
    }
}

sub _process_name {
    my ($self, $name)=@_;
    my $pkgpair=$self->_match_and_filter($name);
    return unless $pkgpair;
    $self->_process_pair_wrapper($pkgpair);
}

sub _process_transaction {
    my ($self, $transaction_ptr)=@_;
    my $transaction_name=$transaction_ptr->[0];
    my $is_in_transaction=0;
    my $transaction_count=0;
    foreach my $name (@$transaction_ptr) {
	my $pkgpair=$self->_match_and_filter($name);
	next if not $pkgpair;
	if (not $is_in_transaction) {
	    $self->init_transaction($transaction_name);
	    $is_in_transaction=1;
	}
	$self->_process_pair_wrapper($pkgpair, $transaction_name, ++$transaction_count);
    }
    $self->done_transaction($transaction_name,$transaction_count) if $is_in_transaction;
}

sub _match_and_filter {
    my ($self,$input)=@_;
    Carp::confess "Source::Repository::Mass::new_pair: undefined input" unless defined $input;
    my $resourceobj=Source::Shared::ResourceContainer::get_resourceobj();

    my ($destname, $originname, $dest, $origin);
    my $originrepo=$self->{-originrepo};
    my $destrepo=$self->{-destrepo};

    if ($self->{-destmode}) {
	# if (-f $input && $destrepo && $destrepo->is_pkg($input)) {
	#     $dest=$destrepo->file2pkg($input);
	#     $destname=$dest->name;
	# } else {
	    $destname=$input;
	# }
    } else {
	if (-f $input && $originrepo && $originrepo->is_pkg($input)) {
	    $origin=$originrepo->file2pkg($input);
	    $originname=$origin->name;
	} else {
	    $originname=$input;
	}
    }
    if ($self->{-destmode}) {
	$originname=$self->_get_matcher()->dest2origin($destname,$resourceobj);
    } else {
	$destname=$self->_get_matcher()->origin2dest($originname,$resourceobj);
    }

    if ($verbose > 2) {
	print STDERR "Source::Repository::Matcher: new_pair($input): origin=$originname dest=$destname\n";
	print STDERR "Source::Repository::Matcher: originrepo=".ref($originrepo)." destrepo=".ref($destrepo)."\n";
	print STDERR "Source::Repository::Matcher: origin=".ref($origin)." dest=".ref($dest)."\n";
    }

    if (Source::Repository::Mass::NameFilter->namefilter($originname,$destname,$resourceobj)) {
	print STDERR "Source::Repository::Matcher: ($originname,$destname) rejected by namefilter\n" if $verbose >1;
	return;
    } else {
	Source::Repository::Mass::NameFilter->postnamefilter($originname,$destname,$resourceobj);
    }

    my @pair_args=(
	-originrepo => $originrepo,
	-destrepo => $destrepo,
	-originname => $originname,
	-destname => $destname,
	);
    push @pair_args, -origin => $origin if $origin;
    push @pair_args, -dest => $dest if $dest;
    my $pkgpair=Source::Package::Pair->new(@pair_args);

    eval {

	# TODO: put it into Mass::PkgFilter.
	# Mass::PkgFilter should be priority container as is_strategy_allow is rather heavy
	# - should not be first. priorities should be: PREORIGIN PREDEST ORIGIN DEST
	unless ($self->is_strategy_allow($pkgpair)) {
	    print STDERR "Source::Repository::Matcher: ($originname,$destname) rejected by strategy\n" if $verbose >1;
	    $pkgpair=undef;
	    return; # from eval
	}
	if (Source::Repository::Mass::PkgFilter->pkgfilter($pkgpair,$resourceobj,$self)) {
	    print STDERR "Source::Repository::Matcher: ($originname,$destname) rejected by pkgfilter\n" if $verbose >1;
	    $pkgpair=undef;
	    return;  # from eval
	} else {
	    Source::Repository::Mass::PkgFilter->postpkgfilter($pkgpair,$resourceobj,$self);
	}
    };
    if ($@) {
	print STDERR 'FAILED match_and_filter (Mass): ',$pkgpair->originname,": ",$@;
	die "

You can ignore this error with --skip-on-error option\n" unless $opt_skip_on_error;
	return undef;
    }
    return $pkgpair;
}

sub _get_matcher {
    my $self=shift;
    my $matcher_obj=$self->{-matcher};
    unless ($matcher_obj) {
	$matcher_obj = $matcher_class->new(
	-originrepo => $self->{-originrepo}, 
	-destrepo => $self->{-destrepo}, 
	-destmode=>$opt_destmode,
	);
	$self->{-matcher}=$matcher_obj;
    }
    return $matcher_obj;
}

sub init_transaction {
    my ($self,$transaction_name)=@_;
}

sub done_transaction {
    my ($self,$transaction_name)=@_;
}

sub _process_pair_wrapper {
    my ($self,$pkgpair,$transaction_name,$transaction_count)=@_;
    eval {
	$self->process_pair($pkgpair,$transaction_name,$transaction_count);
    };
    if ($@) {
	die 'FAILED (Mass): '.$pkgpair->originname.": $@

You can ignore this error with --skip-on-error option\n" unless $opt_skip_on_error;
	print STDERR 'FAILED (Mass): '.$pkgpair->originname.": $@";
    };
}

###############################################################
# The method process_pair should be implemented in subclasses;
# disable the prototype implementation completely to allow 
# multiple inheritance tricks with Transform::{Embedded,PassThrough}
###############################################################
#sub process_pair {
#    my ($self,$pkgpair,$transaction_name,$transaction_count)=@_;
#    print STDERR "process_pair: $self\n";
#    ...
#}

sub is_strategy_allow {
    my ($self,$pkgpair)=@_;
    my $strategy_allow=$STRATEGYMAP{$strategy};
    unless ($strategy_allow) {
	&_print_strategies();
	die "strategy $strategy not found.\n" ;
    }
    ## TODO: ABSTRACTIZE AS FILTERS
    my $originname=$pkgpair->originname;
    my $destname=$pkgpair->destname;
    return 1 if &{$strategy_allow}($self, $pkgpair->origin, $pkgpair->dest, $originname, $destname);
    return 0;
}

#------------------------------------------

#------------------------------------------
#    private strategies
#------------------------------------------

sub help_strategy {
    &_print_strategies();
    exit(1);
}

sub origin2dest {
    my ($self,$originname, $destname,$origin, $dest)=@_;
    print STDERR "warning: origin PKG $originname not found.\n" unless $origin;
    return 0 unless $origin;
    print STDERR "note: dest PKG $destname not found.\n" unless $dest;
    return 1;
}

sub origin2destS {
    my ($self,$originname, $destname,$origin, $dest)=@_;
    #print STDERR "warning: origin PKG $originname not found.\n" unless $origin;
    return 0 unless $origin;
    #print STDERR "note: dest PKG $destname not found.\n" unless $dest;
    return 1;
}

sub destup {
    my ($self, $origin, $dest, $originname, $destname)=@_;
    print STDERR "warning: origin PKG $originname not found.\n" unless $origin;
    return 0 unless $origin;
    print STDERR "note: dest PKG $destname not found.\n" unless $dest;
    return 1 if $dest and $self->comparator->cmp($origin,$dest)>0;
    return 0;
}

sub destupS {
    my ($self, $origin, $dest, $originname, $destname)=@_;
    #print STDERR "warning: origin PKG $originname not found.\n" unless $origin;
    return 0 unless $origin;
    #print STDERR "note: dest PKG $destname not found.\n" unless $dest;
    return 1 if $dest and $self->comparator->cmp($origin,$dest)>0;
    return 0;
}

sub destnotnewer {
    my ($self, $origin, $dest, $originname, $destname)=@_;
    print STDERR "warning: origin PKG $originname not found.\n" unless $origin;
    return 0 unless $origin;
    #print STDERR "note: dest PKG $destname not found.\n" unless $dest;
    return 1 if $dest and $self->comparator->cmp($origin,$dest)>=0;
    return 0;
}

sub destnotnewerS {
    my ($self, $origin, $dest, $originname, $destname)=@_;
    #print STDERR "warning: origin PKG $originname not found.\n" unless $origin;
    return 0 unless $origin;
    #print STDERR "note: dest PKG $destname not found.\n" unless $dest;
    return 1 if $dest and $self->comparator->cmp($origin,$dest)>=0;
    return 0;
}

sub newer {
    my ($self, $origin, $dest, $originname, $destname)=@_;
    print STDERR "warning: origin PKG $originname not found.\n" unless $origin;
    return 0 unless $origin;
    #print STDERR "note: dest PKG $originname not found.\n" unless $dest;
    return 1 if not $dest or $self->comparator->cmp($origin,$dest)>0;
    return 0;
}

sub newerS {
    my ($self, $origin, $dest, $originname, $destname)=@_;
    #print STDERR "warning: origin PKG $originname not found.\n" unless $origin;
    return 0 unless $origin;
    #print STDERR "note: dest PKG $originname not found.\n" unless $dest;
    return 1 if not $dest or $self->comparator->cmp($origin,$dest)>0;
    return 0;
}

sub both {
    my ($self, $origin, $dest, $originname, $destname)=@_;
    print STDERR "warning: origin PKG $originname not found.\n" unless $origin;
    return 0 unless $origin;
    #print STDERR "note: dest PKG $originname not found.\n" unless $dest;
    return 1 if $dest;
    return 0;
}


sub bothS {
    my ($self, $origin, $dest, $originname, $destname)=@_;
    #print STDERR "warning: origin PKG $originname not found.\n" unless $origin;
    return 0 unless $origin;
    #print STDERR "note: dest PKG $originname not found.\n" unless $dest;
    return 1 if $dest;
    return 0;
}

sub originskipdest {
    my ($self, $origin, $dest, $originname, $destname)=@_;
    print STDERR "warning: origin PKG $originname not found.\n" unless $origin;
    return 0 unless $origin;
    print STDERR "found dest PKG $destname for $originname.\n" if $dest;
    return 0 if $dest;
    return 1;
}

sub originskipdestS {
    my ($self, $origin, $dest, $originname, $destname)=@_;
    #print STDERR "warning: origin PKG $originname not found.\n" unless $origin;
    return 0 unless $origin;
    #print STDERR "found dest PKG $destname for $originname.\n" if $dest;
    return 0 if $dest;
    return 1;
}

sub newolder {
    my ($self, $origin, $dest, $originname, $destname)=@_;
    #print STDERR "warning: origin PKG $originname not found.\n" unless $origin;
    return 1 unless $origin;
    #print STDERR "note: dest PKG $destname not found.\n" unless $dest;
    return 1 if $dest and $self->comparator->cmp($origin,$dest)<0;
    return 0;
}

sub notolder {
    my ($self, $origin, $dest, $originname, $destname)=@_;
    print STDERR "warning: origin PKG $originname not found.\n" unless $origin;
    return 0 unless $origin;
    return 1 unless $dest;
    return 1 if $self->comparator->cmp($origin,$dest)>=0;
    return 0;
}

sub older {
    my ($self, $origin, $dest, $originname, $destname)=@_;
    print STDERR "warning: origin PKG $originname not found.\n" unless $origin;
    return 0 unless $origin;
    #print STDERR "note: dest PKG $destname not found.\n" unless $dest;
    return 1 if $dest and $self->comparator->cmp($origin,$dest)<0;
    return 0;
}

sub olderS {
    my ($self, $origin, $dest, $originname, $destname)=@_;
    #print STDERR "warning: origin PKG $originname not found.\n" unless $origin;
    return 0 unless $origin;
    #print STDERR "note: dest PKG $destname not found.\n" unless $dest;
    return 1 if $dest and $self->comparator->cmp($origin,$dest)<0;
    return 0;
}

sub destskiporigin {
    my ($self, $origin, $dest, $originname, $destname)=@_;
    print STDERR "warning: dest PKG $destname not found.\n" unless $dest;
    return 0 unless $dest;
    print STDERR "found origin PKG $originname for $destname.\n" if $origin;
    return 0 if $origin;
    return 1;
}

sub destskiporiginS {
    my ($self, $origin, $dest, $originname, $destname)=@_;
    #print STDERR "warning: dest PKG $destname not found.\n" unless $dest;
    return 0 unless $dest;
    #print STDERR "found origin PKG $originname for $destname.\n" if $origin;
    return 0 if $origin;
    return 1;
}

## ########
# release up; version is the same, something other (release) is different
## ########
sub relup {
    my ($self, $origin, $dest, $originname, $destname)=@_;
    #print STDERR "warning: origin PKG $originname not found.\n" unless $origin;
    return 0 unless $origin;
    #print STDERR "note: dest PKG $destname not found.\n" unless $dest;
    return 1 if $dest and $dest->version eq $origin->version and $self->comparator->cmp($origin,$dest)>0;
    return 0;
}

## ########
# version up; versions are different
## ########
sub verup {
    my ($self, $origin, $dest, $originname, $destname)=@_;
    #print STDERR "warning: origin PKG $originname not found.\n" unless $origin;
    return 0 unless $origin;
    #print STDERR "note: dest PKG $destname not found.\n" unless $dest;
    return 1 if $dest and $dest->version ne $origin->version and $self->comparator->cmp($origin,$dest)>0;
    return 0;
}

sub same {
    my ($self, $origin, $dest, $originname, $destname)=@_;
    return 0 unless $origin and $dest;
    return 1 if $self->comparator->cmp($origin,$dest)==0;
    return 0;
}

#------------------------------------------

1;
